@@ -1,11 +1,14 @@
 # -*- encoding:utf-8 -*-
-from __future__ import division, absolute_import, print_function
-
-import sys
+from collections import namedtuple
+from copy import deepcopy
+import re
 import textwrap
+import warnings
 
 import jinja2
 
+from numpydoc.numpydoc import update_config
+from numpydoc.xref import DEFAULT_LINKS
 from numpydoc.docscrape import (
     NumpyDocString,
     FunctionDoc,
@@ -13,14 +16,10 @@ from numpydoc.docscrape import (
     ParseError
 )
 from numpydoc.docscrape_sphinx import (SphinxDocString, SphinxClassDoc,
-                                       SphinxFunctionDoc)
-from nose.tools import (assert_equal, assert_raises, assert_list_equal,
-                        assert_true)
-
-if sys.version_info[0] >= 3:
-    sixu = lambda s: s
-else:
-    sixu = lambda s: unicode(s, 'unicode_escape')
+                                       SphinxFunctionDoc, get_doc_object)
+import pytest
+from pytest import raises as assert_raises
+from pytest import warns as assert_warns
 
 
 doc_txt = '''\
@@ -47,12 +46,14 @@ doc_txt = '''\
       Given a shape of, for example, (m,n,k), m*n*k samples are
       generated, and packed in an m-by-n-by-k arrangement.  Because
       each sample is N-dimensional, the output shape is (m,n,k,N).
+  dtype : data type object, optional (default : float)
+      The type and size of the data to be returned.
 
   Returns
   -------
   out : ndarray
       The drawn samples, arranged according to `shape`.  If the
-      shape given is (m,n,...), then the shape of `out` is is
+      shape given is (m,n,...), then the shape of `out` is
       (m,n,...,N).
 
       In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
@@ -60,6 +61,7 @@ doc_txt = '''\
   list of str
       This is not a real return value.  It exists to test
       anonymous return values.
+  no_description
 
   Other Parameters
   ----------------
@@ -111,26 +113,31 @@ doc_txt = '''\
   --------
   some, other, funcs
   otherfunc : relationship
+  :py:meth:`spyder.widgets.mixins.GetHelpMixin.show_object_info`
 
   Examples
   --------
   >>> mean = (1,2)
   >>> cov = [[1,0],[1,0]]
   >>> x = multivariate_normal(mean,cov,(3,3))
-  >>> print x.shape
+  >>> print(x.shape)
   (3, 3, 2)
 
   The following is probably true, given that 0.6 is roughly twice the
   standard deviation:
 
-  >>> print list( (x[0,0,:] - mean) < 0.6 )
+  >>> print(list((x[0, 0, :] - mean) < 0.6))
   [True, True]
 
   .. index:: random
      :refguide: random;distributions, random;gauss
 
   '''
-doc = NumpyDocString(doc_txt)
+
+@pytest.fixture(params=['','\n    '], ids=["flush", "newline_indented"])
+def doc(request):
+    return NumpyDocString(request.param+doc_txt)
+
 
 doc_yields_txt = """
 Test generator
@@ -147,59 +154,115 @@ int
 doc_yields = NumpyDocString(doc_yields_txt)
 
 
-def test_signature():
+doc_sent_txt = """
+Test generator
+
+Yields
+------
+a : int
+    The number of apples.
+
+Receives
+--------
+b : int
+    The number of bananas.
+c : int
+    The number of oranges.
+
+"""
+doc_sent = NumpyDocString(doc_sent_txt)
+
+
+def test_signature(doc):
     assert doc['Signature'].startswith('numpy.multivariate_normal(')
     assert doc['Signature'].endswith('spam=None)')
 
-def test_summary():
+
+def test_summary(doc):
     assert doc['Summary'][0].startswith('Draw values')
     assert doc['Summary'][-1].endswith('covariance.')
 
-def test_extended_summary():
+
+def test_extended_summary(doc):
     assert doc['Extended Summary'][0].startswith('The multivariate normal')
 
-def test_parameters():
-    assert_equal(len(doc['Parameters']), 3)
-    assert_equal([n for n,_,_ in doc['Parameters']], ['mean','cov','shape'])
+
+def test_parameters(doc):
+    assert len(doc['Parameters']) == 4
+    names = [n for n, _, _ in doc['Parameters']]
+    assert all(a == b for a, b in zip(names, ['mean', 'cov', 'shape']))
 
     arg, arg_type, desc = doc['Parameters'][1]
-    assert_equal(arg_type, '(N, N) ndarray')
+    assert arg_type == '(N, N) ndarray'
     assert desc[0].startswith('Covariance matrix')
-    assert doc['Parameters'][0][-1][-2] == '   (1+2+3)/3'
+    assert doc['Parameters'][0][-1][-1] == '   (1+2+3)/3'
 
-def test_other_parameters():
-    assert_equal(len(doc['Other Parameters']), 1)
-    assert_equal([n for n,_,_ in doc['Other Parameters']], ['spam'])
+    arg, arg_type, desc = doc['Parameters'][2]
+    assert arg == 'shape'
+    assert arg_type == 'tuple of ints'
+    assert desc[0].startswith('Given')
+    assert doc['Parameters'][0][-1][-1] == '   (1+2+3)/3'
+
+    arg, arg_type, desc = doc['Parameters'][3]
+    assert arg == 'dtype'
+    assert arg_type == 'data type object, optional (default : float)'
+    assert desc[0].startswith('The type and size')
+
+
+def test_other_parameters(doc):
+    assert len(doc['Other Parameters']) == 1
+    assert [n for n, _, _ in doc['Other Parameters']] == ['spam']
     arg, arg_type, desc = doc['Other Parameters'][0]
-    assert_equal(arg_type, 'parrot')
+    assert arg_type == 'parrot'
     assert desc[0].startswith('A parrot off its mortal coil')
 
-def test_returns():
-    assert_equal(len(doc['Returns']), 2)
+
+
+def test_returns(doc):
+    assert len(doc['Returns']) == 3
     arg, arg_type, desc = doc['Returns'][0]
-    assert_equal(arg, 'out')
-    assert_equal(arg_type, 'ndarray')
+    assert arg == 'out'
+    assert arg_type == 'ndarray'
     assert desc[0].startswith('The drawn samples')
     assert desc[-1].endswith('distribution.')
 
     arg, arg_type, desc = doc['Returns'][1]
-    assert_equal(arg, 'list of str')
-    assert_equal(arg_type, '')
+    assert arg == ''
+    assert arg_type == 'list of str'
     assert desc[0].startswith('This is not a real')
     assert desc[-1].endswith('anonymous return values.')
 
+    arg, arg_type, desc = doc['Returns'][2]
+    assert arg == ''
+    assert arg_type == 'no_description'
+    assert not ''.join(desc).strip()
+
+
 def test_yields():
     section = doc_yields['Yields']
-    assert_equal(len(section), 3)
+    assert len(section) == 3
     truth = [('a', 'int', 'apples.'),
              ('b', 'int', 'bananas.'),
-             ('int', '', 'unknowns.')]
+             ('', 'int', 'unknowns.')]
     for (arg, arg_type, desc), (arg_, arg_type_, end) in zip(section, truth):
-        assert_equal(arg, arg_)
-        assert_equal(arg_type, arg_type_)
+        assert arg == arg_
+        assert arg_type == arg_type_
         assert desc[0].startswith('The number of')
         assert desc[0].endswith(end)
 
+
+def test_sent():
+    section = doc_sent['Receives']
+    assert len(section) == 2
+    truth = [('b', 'int', 'bananas.'),
+             ('c', 'int', 'oranges.')]
+    for (arg, arg_type, desc), (arg_, arg_type_, end) in zip(section, truth):
+        assert arg == arg_
+        assert arg_type == arg_type_
+        assert desc[0].startswith('The number of')
+        assert desc[0].endswith(end)
+
+
 def test_returnyield():
     doc_text = """
 Test having returns and yields.
@@ -232,10 +295,11 @@ Notes
 -----
 That should break...
 """
-    assert_raises(ValueError, NumpyDocString, doc_text)
+    with pytest.raises(ValueError, match="The section Notes appears twice"):
+        NumpyDocString(doc_text)
 
     # if we have a numpydoc object, we know where the error came from
-    class Dummy(object):
+    class Dummy:
         """
         Dummy class.
 
@@ -269,51 +333,58 @@ That should break...
         Second note.
         """
 
-    try:
+    with pytest.raises(ValueError, match="Dummy class"):
         SphinxClassDoc(Dummy)
-    except ValueError as e:
-        # python 3 version or python 2 version
-        assert_true("test_section_twice.<locals>.Dummy" in str(e)
-                    or 'test_docscrape.Dummy' in str(e))
 
-    try:
+    with pytest.raises(ValueError, match="dummy_func"):
         SphinxFunctionDoc(dummy_func)
-    except ValueError as e:
-        # python 3 version or python 2 version
-        assert_true("test_section_twice.<locals>.dummy_func" in str(e)
-                    or 'function dummy_func' in str(e))
 
 
-def test_notes():
+def test_notes(doc):
     assert doc['Notes'][0].startswith('Instead')
     assert doc['Notes'][-1].endswith('definite.')
-    assert_equal(len(doc['Notes']), 17)
+    assert len(doc['Notes']) == 17
 
-def test_references():
+
+def test_references(doc):
     assert doc['References'][0].startswith('..')
     assert doc['References'][-1].endswith('2001.')
 
-def test_examples():
+
+def test_examples(doc):
     assert doc['Examples'][0].startswith('>>>')
     assert doc['Examples'][-1].endswith('True]')
 
-def test_index():
-    assert_equal(doc['index']['default'], 'random')
-    assert_equal(len(doc['index']), 2)
-    assert_equal(len(doc['index']['refguide']), 2)
 
-def non_blank_line_by_line_compare(a,b):
+def test_index(doc):
+    assert doc['index']['default'] == 'random'
+    assert len(doc['index']) == 2
+    assert len(doc['index']['refguide']) == 2
+
+
+def _strip_blank_lines(s):
+    "Remove leading, trailing and multiple blank lines"
+    s = re.sub(r'^\s*\n', '', s)
+    s = re.sub(r'\n\s*$', '', s)
+    s = re.sub(r'\n\s*\n', r'\n\n', s)
+    return s
+
+
+def line_by_line_compare(a, b, n_lines=None):
     a = textwrap.dedent(a)
     b = textwrap.dedent(b)
-    a = [l.rstrip() for l in a.split('\n') if l.strip()]
-    b = [l.rstrip() for l in b.split('\n') if l.strip()]
-    assert_list_equal(a, b)
+    a = [l.rstrip() for l in _strip_blank_lines(a).split('\n')][:n_lines]
+    b = [l.rstrip() for l in _strip_blank_lines(b).split('\n')][:n_lines]
+    assert len(a) == len(b)
+    for ii, (aa, bb) in enumerate(zip(a, b)):
+        assert aa == bb
 
-def test_str():
+
+def test_str(doc):
     # doc_txt has the order of Notes and See Also sections flipped.
     # This should be handled automatically, and so, one thing this test does
     # is to make sure that See Also precedes Notes in the output.
-    non_blank_line_by_line_compare(str(doc),
+    line_by_line_compare(str(doc),
 """numpy.multivariate_normal(mean, cov, shape=None, spam=None)
 
 Draw values from a multivariate normal distribution with specified
@@ -330,19 +401,20 @@ mean : (N,) ndarray
     .. math::
 
        (1+2+3)/3
-
 cov : (N, N) ndarray
     Covariance matrix of the distribution.
 shape : tuple of ints
     Given a shape of, for example, (m,n,k), m*n*k samples are
     generated, and packed in an m-by-n-by-k arrangement.  Because
     each sample is N-dimensional, the output shape is (m,n,k,N).
+dtype : data type object, optional (default : float)
+    The type and size of the data to be returned.
 
 Returns
 -------
 out : ndarray
     The drawn samples, arranged according to `shape`.  If the
-    shape given is (m,n,...), then the shape of `out` is is
+    shape given is (m,n,...), then the shape of `out` is
     (m,n,...,N).
 
     In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
@@ -350,6 +422,7 @@ out : ndarray
 list of str
     This is not a real return value.  It exists to test
     anonymous return values.
+no_description
 
 Other Parameters
 ----------------
@@ -372,10 +445,13 @@ Certain warnings apply.
 
 See Also
 --------
-`some`_, `other`_, `funcs`_
 
+`some`_, `other`_, `funcs`_
+    ..
 `otherfunc`_
     relationship
+:py:meth:`spyder.widgets.mixins.GetHelpMixin.show_object_info`
+    ..
 
 Notes
 -----
@@ -409,13 +485,13 @@ Examples
 >>> mean = (1,2)
 >>> cov = [[1,0],[1,0]]
 >>> x = multivariate_normal(mean,cov,(3,3))
->>> print x.shape
+>>> print(x.shape)
 (3, 3, 2)
 
 The following is probably true, given that 0.6 is roughly twice the
 standard deviation:
 
->>> print list( (x[0,0,:] - mean) < 0.6 )
+>>> print(list((x[0, 0, :] - mean) < 0.6))
 [True, True]
 
 .. index:: random
@@ -423,7 +499,7 @@ standard deviation:
 
 
 def test_yield_str():
-    non_blank_line_by_line_compare(str(doc_yields),
+    line_by_line_compare(str(doc_yields),
 """Test generator
 
 Yields
@@ -434,13 +510,46 @@ b : int
     The number of bananas.
 int
     The number of unknowns.
+""")
 
-.. index:: """)
 
+def test_receives_str():
+    line_by_line_compare(str(doc_sent),
+"""Test generator
+
+Yields
+------
+a : int
+    The number of apples.
+
+Receives
+--------
+b : int
+    The number of bananas.
+c : int
+    The number of oranges.
+""")
+
+
+def test_no_index_in_str():
+    assert "index" not in str(NumpyDocString("""Test idx
+
+    """))
+
+    assert "index" in str(NumpyDocString("""Test idx
+
+    .. index :: random
+    """))
+
+    assert "index" in str(NumpyDocString("""Test idx
+
+    .. index ::
+        foo
+    """))
 
 def test_sphinx_str():
     sphinx_doc = SphinxDocString(doc_txt)
-    non_blank_line_by_line_compare(str(sphinx_doc),
+    line_by_line_compare(str(sphinx_doc),
 """
 .. index:: random
    single: random;distributions, random;gauss
@@ -454,7 +563,6 @@ of the one-dimensional normal distribution to higher dimensions.
 :Parameters:
 
     **mean** : (N,) ndarray
-
         Mean of the N-dimensional distribution.
 
         .. math::
@@ -462,47 +570,46 @@ of the one-dimensional normal distribution to higher dimensions.
            (1+2+3)/3
 
     **cov** : (N, N) ndarray
-
         Covariance matrix of the distribution.
 
     **shape** : tuple of ints
-
         Given a shape of, for example, (m,n,k), m*n*k samples are
         generated, and packed in an m-by-n-by-k arrangement.  Because
         each sample is N-dimensional, the output shape is (m,n,k,N).
 
+    **dtype** : data type object, optional (default : float)
+        The type and size of the data to be returned.
+
 :Returns:
 
     **out** : ndarray
-
         The drawn samples, arranged according to `shape`.  If the
-        shape given is (m,n,...), then the shape of `out` is is
+        shape given is (m,n,...), then the shape of `out` is
         (m,n,...,N).
 
         In other words, each entry ``out[i,j,...,:]`` is an N-dimensional
         value drawn from the distribution.
 
     list of str
-
         This is not a real return value.  It exists to test
         anonymous return values.
 
+    no_description
+        ..
+
 :Other Parameters:
 
     **spam** : parrot
-
         A parrot off its mortal coil.
 
 :Raises:
 
-    **RuntimeError**
-
+    RuntimeError
         Some error
 
 :Warns:
 
-    **RuntimeWarning**
-
+    RuntimeWarning
         Some warning
 
 .. warning::
@@ -512,9 +619,11 @@ of the one-dimensional normal distribution to higher dimensions.
 .. seealso::
 
     :obj:`some`, :obj:`other`, :obj:`funcs`
-
+        ..
     :obj:`otherfunc`
         relationship
+    :py:meth:`spyder.widgets.mixins.GetHelpMixin.show_object_info`
+        ..
 
 .. rubric:: Notes
 
@@ -552,34 +661,31 @@ definite.
 >>> mean = (1,2)
 >>> cov = [[1,0],[1,0]]
 >>> x = multivariate_normal(mean,cov,(3,3))
->>> print x.shape
+>>> print(x.shape)
 (3, 3, 2)
 
 The following is probably true, given that 0.6 is roughly twice the
 standard deviation:
 
->>> print list( (x[0,0,:] - mean) < 0.6 )
+>>> print(list((x[0, 0, :] - mean) < 0.6))
 [True, True]
 """)
 
 
 def test_sphinx_yields_str():
     sphinx_doc = SphinxDocString(doc_yields_txt)
-    non_blank_line_by_line_compare(str(sphinx_doc),
+    line_by_line_compare(str(sphinx_doc),
 """Test generator
 
 :Yields:
 
     **a** : int
-
         The number of apples.
 
     **b** : int
-
         The number of bananas.
 
     int
-
         The number of unknowns.
 """)
 
@@ -595,8 +701,10 @@ doc2 = NumpyDocString("""
         If None, the index is into the flattened array, otherwise along
         the specified axis""")
 
+
 def test_parameters_without_extended_description():
-    assert_equal(len(doc2['Parameters']), 2)
+    assert len(doc2['Parameters']) == 2
+
 
 doc3 = NumpyDocString("""
     my_signature(*params, **kwds)
@@ -604,23 +712,27 @@ doc3 = NumpyDocString("""
     Return this and that.
     """)
 
+
 def test_escape_stars():
     signature = str(doc3).split('\n')[0]
-    assert_equal(signature, 'my_signature(\*params, \*\*kwds)')
+    assert signature == r'my_signature(\*params, \*\*kwds)'
 
     def my_func(a, b, **kwargs):
         pass
 
     fdoc = FunctionDoc(func=my_func)
-    assert_equal(fdoc['Signature'], 'my_func(a, b, \*\*kwargs)')
+    assert fdoc['Signature'] == ''
+
 
 doc4 = NumpyDocString(
     """a.conj()
 
     Return an array with all complex-valued elements conjugated.""")
 
+
 def test_empty_extended_summary():
-    assert_equal(doc4['Extended Summary'], [])
+    assert doc4['Extended Summary'] == []
+
 
 doc5 = NumpyDocString(
     """
@@ -637,22 +749,32 @@ doc5 = NumpyDocString(
         If needed
     """)
 
+
 def test_raises():
-    assert_equal(len(doc5['Raises']), 1)
-    name,_,desc = doc5['Raises'][0]
-    assert_equal(name,'LinAlgException')
-    assert_equal(desc,['If array is singular.'])
+    assert len(doc5['Raises']) == 1
+    param = doc5['Raises'][0]
+    assert param.name == ''
+    assert param.type == 'LinAlgException'
+    assert param.desc == ['If array is singular.']
+
 
 def test_warns():
-    assert_equal(len(doc5['Warns']), 1)
-    name,_,desc = doc5['Warns'][0]
-    assert_equal(name,'SomeWarning')
-    assert_equal(desc,['If needed'])
+    assert len(doc5['Warns']) == 1
+    param = doc5['Warns'][0]
+    assert param.name == ''
+    assert param.type == 'SomeWarning'
+    assert param.desc == ['If needed']
 
-def test_see_also():
+# see numpydoc/numpydoc #281
+# we want to correctly parse "See Also" both in docstrings both like
+#"""foo
+# and
+#"""
+#foo
+@pytest.mark.parametrize('prefix', ['', '\n    '])
+def test_see_also(prefix):
     doc6 = NumpyDocString(
-    """
-    z(x,theta)
+    prefix + """z(x,theta)
 
     See Also
     --------
@@ -662,34 +784,46 @@ def test_see_also():
              multiple lines
     func_f, func_g, :meth:`func_h`, func_j,
     func_k
+    func_f1, func_g1, :meth:`func_h1`, func_j1
+    func_f2, func_g2, :meth:`func_h2`, func_j2 : description of multiple
     :obj:`baz.obj_q`
+    :obj:`~baz.obj_r`
     :class:`class_j`: fubar
         foobar
     """)
 
-    assert len(doc6['See Also']) == 12
-    for func, desc, role in doc6['See Also']:
-        if func in ('func_a', 'func_b', 'func_c', 'func_f',
-                    'func_g', 'func_h', 'func_j', 'func_k', 'baz.obj_q'):
-            assert(not desc)
-        else:
-            assert(desc)
+    assert len(doc6['See Also']) == 10
+    for funcs, desc in doc6['See Also']:
+        for func, role in funcs:
+            if func in ('func_a', 'func_b', 'func_c', 'func_f',
+                        'func_g', 'func_h', 'func_j', 'func_k', 'baz.obj_q',
+                        'func_f1', 'func_g1', 'func_h1', 'func_j1',
+                        '~baz.obj_r'):
+                assert not desc, str([func, desc])
+            elif func in ('func_f2', 'func_g2', 'func_h2', 'func_j2'):
+                assert desc, str([func, desc])
+            else:
+                assert desc, str([func, desc])
 
-        if func == 'func_h':
-            assert role == 'meth'
-        elif func == 'baz.obj_q':
-            assert role == 'obj'
-        elif func == 'class_j':
-            assert role == 'class'
-        else:
-            assert role is None
+            if func == 'func_h':
+                assert role == 'meth'
+            elif func == 'baz.obj_q' or func == '~baz.obj_r':
+                assert role == 'obj'
+            elif func == 'class_j':
+                assert role == 'class'
+            elif func in ['func_h1', 'func_h2']:
+                assert role == 'meth'
+            else:
+                assert role is None, str([func, role])
 
-        if func == 'func_d':
-            assert desc == ['some equivalent func']
-        elif func == 'foo.func_e':
-            assert desc == ['some other func over', 'multiple lines']
-        elif func == 'class_j':
-            assert desc == ['fubar', 'foobar']
+            if func == 'func_d':
+                assert desc == ['some equivalent func']
+            elif func == 'foo.func_e':
+                assert desc == ['some other func over', 'multiple lines']
+            elif func == 'class_j':
+                assert desc == ['fubar', 'foobar']
+            elif func in ['func_f2', 'func_g2', 'func_h2', 'func_j2']:
+                assert desc == ['description of multiple'], str([desc, ['description of multiple']])
 
 
 def test_see_also_parse_error():
@@ -701,15 +835,12 @@ def test_see_also_parse_error():
     --------
     :func:`~foo`
     """)
-    with assert_raises(ParseError) as err:
+    with pytest.raises(ValueError, match="See Also entry ':func:`~foo`'"):
         NumpyDocString(text)
-    assert_equal(
-        str(r":func:`~foo` is not a item name in '\n    z(x,theta)\n\n    See Also\n    --------\n    :func:`~foo`\n    '"),
-        str(err.exception)
-    )
+
 
 def test_see_also_print():
-    class Dummy(object):
+    class Dummy:
         """
         See Also
         --------
@@ -720,21 +851,82 @@ def test_see_also_print():
         """
         pass
 
-    obj = Dummy()
-    s = str(FunctionDoc(obj, role='func'))
+    s = str(FunctionDoc(Dummy, role='func'))
     assert(':func:`func_a`, :func:`func_b`' in s)
     assert('    some relationship' in s)
     assert(':func:`func_d`' in s)
 
+
+def test_see_also_trailing_comma_warning():
+    warnings.filterwarnings('error')
+    with assert_warns(Warning, match='Unexpected comma or period after function list at index 43 of line .*'):
+        NumpyDocString(
+            """
+            z(x,theta)
+
+            See Also
+            --------
+            func_f2, func_g2, :meth:`func_h2`, func_j2, : description of multiple
+            :class:`class_j`: fubar
+                foobar
+            """)
+
+
+def test_unknown_section():
+    doc_text = """
+Test having an unknown section
+
+Mope
+----
+This should be ignored and warned about
+"""
+
+    class BadSection:
+        """Class with bad section.
+
+        Nope
+        ----
+        This class has a nope section.
+        """
+        pass
+
+    with pytest.warns(UserWarning, match="Unknown section Mope") as record:
+        NumpyDocString(doc_text)
+    assert len(record) == 1
+
+    # SphinxClassDoc has _obj.__name__ == "BadSection". Test that this is
+    # included in the message
+    msg_match = "Unknown section Nope in the docstring of BadSection"
+    with pytest.warns(UserWarning, match=msg_match) as record:
+        SphinxClassDoc(BadSection)
+    assert len(record) == 1
+
+
 doc7 = NumpyDocString("""
 
         Doc starts on second line.
 
         """)
 
+
 def test_empty_first_line():
     assert doc7['Summary'][0].startswith('Doc starts')
 
+doc8 = NumpyDocString("""
+
+        Parameters with colon and no types:
+
+        Parameters
+        ----------
+
+        data :
+            some stuff, technically invalid
+        """)
+
+
+def test_trailing_colon():
+    assert doc8['Parameters'][0].name == 'data'
+
 
 def test_no_summary():
     str(SphinxDocString("""
@@ -762,6 +954,7 @@ def test_unicode():
     assert isinstance(doc['Summary'][0], str)
     assert doc['Summary'][0] == 'öäöäöäöäöåååå'
 
+
 def test_plot_examples():
     cfg = dict(use_plots=True)
 
@@ -774,6 +967,15 @@ def test_plot_examples():
     """, config=cfg)
     assert 'plot::' in str(doc), str(doc)
 
+    doc = SphinxDocString("""
+    Examples
+    --------
+    >>> from matplotlib import pyplot as plt
+    >>> plt.plot([1,2,3],[4,5,6])
+    >>> plt.show()
+    """, config=cfg)
+    assert 'plot::' in str(doc), str(doc)
+
     doc = SphinxDocString("""
     Examples
     --------
@@ -785,9 +987,50 @@ def test_plot_examples():
     """, config=cfg)
     assert str(doc).count('plot::') == 1, str(doc)
 
+
+def test_use_blockquotes():
+    cfg = dict(use_blockquotes=True)
+    doc = SphinxDocString("""
+    Parameters
+    ----------
+    abc : def
+        ghi
+    jkl
+        mno
+
+    Returns
+    -------
+    ABC : DEF
+        GHI
+    JKL
+        MNO
+    """, config=cfg)
+    line_by_line_compare(str(doc), '''
+    :Parameters:
+
+        **abc** : def
+
+            ghi
+
+        **jkl**
+
+            mno
+
+    :Returns:
+
+        **ABC** : DEF
+
+            GHI
+
+        JKL
+
+            MNO
+    ''')
+
+
 def test_class_members():
 
-    class Dummy(object):
+    class Dummy:
         """
         Dummy class.
 
@@ -803,7 +1046,7 @@ def test_class_members():
             """Spammity index"""
             return 0.95
 
-        class Ignorable(object):
+        class Ignorable:
             """local class, to be ignored"""
             pass
 
@@ -866,6 +1109,7 @@ def test_class_members():
         else:
             assert 'Spammity index' in str(doc), str(doc)
 
+
 def test_duplicate_signature():
     # Duplicate function signatures occur e.g. in ufuncs, when the
     # automatic mechanism adds one, and a more detailed comes from the
@@ -889,6 +1133,7 @@ class_doc_txt = """
     f : callable ``f(t, y, *f_args)``
         Aaa.
     jac : callable ``jac(t, y, *jac_args)``
+
         Bbb.
 
     Attributes
@@ -897,8 +1142,17 @@ class_doc_txt = """
         Current time.
     y : ndarray
         Current variable values.
-    x : float
-        Some parameter
+
+        * hello
+        * world
+    an_attribute : float
+        The docstring is printed instead
+    no_docstring : str
+        But a description
+    no_docstring2 : str
+    multiline_sentence
+    midword_period
+    no_period
 
     Methods
     -------
@@ -911,9 +1165,10 @@ class_doc_txt = """
     For usage examples, see `ode`.
 """
 
+
 def test_class_members_doc():
     doc = ClassDoc(None, class_doc_txt)
-    non_blank_line_by_line_compare(str(doc),
+    line_by_line_compare(str(doc),
     """
     Foo
 
@@ -934,58 +1189,107 @@ def test_class_members_doc():
         Current time.
     y : ndarray
         Current variable values.
-    x : float
-        Some parameter
+
+        * hello
+        * world
+    an_attribute : float
+        The docstring is printed instead
+    no_docstring : str
+        But a description
+    no_docstring2 : str
+    multiline_sentence
+    midword_period
+    no_period
 
     Methods
     -------
     a
-
     b
-
     c
 
-    .. index::
-
     """)
 
+
 def test_class_members_doc_sphinx():
     class Foo:
         @property
-        def x(self):
+        def an_attribute(self):
             """Test attribute"""
             return None
 
+        @property
+        def no_docstring(self):
+            return None
+
+        @property
+        def no_docstring2(self):
+            return None
+
+        @property
+        def multiline_sentence(self):
+            """This is a
+            sentence. It spans multiple lines."""
+            return None
+
+        @property
+        def midword_period(self):
+            """The sentence for numpy.org."""
+            return None
+
+        @property
+        def no_period(self):
+            """This does not have a period
+            so we truncate its summary to the first linebreak
+
+            Apparently.
+            """
+            return None
+
     doc = SphinxClassDoc(Foo, class_doc_txt)
-    non_blank_line_by_line_compare(str(doc),
+    line_by_line_compare(str(doc),
     """
     Foo
 
     :Parameters:
 
         **f** : callable ``f(t, y, *f_args)``
-
             Aaa.
 
         **jac** : callable ``jac(t, y, *jac_args)``
-
             Bbb.
 
     .. rubric:: Examples
 
     For usage examples, see `ode`.
 
-    .. rubric:: Attributes
+    :Attributes:
 
-    .. autosummary::
-       :toctree:
+        **t** : float
+            Current time.
 
-       x
+        **y** : ndarray
+            Current variable values.
 
-    =====  ==========
-    **t**  (float) Current time.
-    **y**  (ndarray) Current variable values.
-    =====  ==========
+            * hello
+            * world
+
+        :obj:`an_attribute <an_attribute>` : float
+            Test attribute
+
+        **no_docstring** : str
+            But a description
+
+        **no_docstring2** : str
+            ..
+
+        :obj:`multiline_sentence <multiline_sentence>`
+            This is a sentence.
+
+        :obj:`midword_period <midword_period>`
+            The sentence for numpy.org.
+
+        :obj:`no_period <no_period>`
+            This does not have a period
 
     .. rubric:: Methods
 
@@ -997,31 +1301,227 @@ def test_class_members_doc_sphinx():
 
     """)
 
+
+def test_class_attributes_as_member_list():
+
+    class Foo:
+        """
+        Class docstring.
+
+        Attributes
+        ----------
+        an_attribute
+            Another description that is not used.
+
+        """
+        @property
+        def an_attribute(self):
+            """Test attribute"""
+            return None
+
+    attr_doc = """:Attributes:
+
+    :obj:`an_attribute <an_attribute>`
+        Test attribute"""
+
+    assert attr_doc in str(SphinxClassDoc(Foo))
+    assert "Another description" not in str(SphinxClassDoc(Foo))
+
+    attr_doc2 = """.. rubric:: Attributes
+
+.. autosummary::
+   :toctree:
+
+   an_attribute"""
+
+    cfg = dict(attributes_as_param_list=False)
+    assert attr_doc2 in str(SphinxClassDoc(Foo, config=cfg))
+    assert "Another description" not in str(SphinxClassDoc(Foo, config=cfg))
+
+
 def test_templated_sections():
     doc = SphinxClassDoc(None, class_doc_txt,
-                         config={'template': jinja2.Template('{{examples}}{{parameters}}')})
-    non_blank_line_by_line_compare(str(doc),
+                         config={'template': jinja2.Template('{{examples}}\n{{parameters}}')})
+    line_by_line_compare(str(doc),
     """
     .. rubric:: Examples
 
     For usage examples, see `ode`.
 
-
     :Parameters:
 
         **f** : callable ``f(t, y, *f_args)``
-
             Aaa.
 
         **jac** : callable ``jac(t, y, *jac_args)``
-
             Bbb.
 
     """)
 
 
+def test_nonstandard_property():
+    # test discovery of a property that does not satisfy isinstace(.., property)
+
+    class SpecialProperty:
+
+        def __init__(self, axis=0, doc=""):
+            self.axis = axis
+            self.__doc__ = doc
+
+        def __get__(self, obj, type):
+            if obj is None:
+                # Only instances have actual _data, not classes
+                return self
+            else:
+                return obj._data.axes[self.axis]
+
+        def __set__(self, obj, value):
+            obj._set_axis(self.axis, value)
+
+    class Dummy:
+
+        attr = SpecialProperty(doc="test attribute")
+
+    doc = get_doc_object(Dummy)
+    assert "test attribute" in str(doc)
+
+
+def test_args_and_kwargs():
+    cfg = dict()
+    doc = SphinxDocString("""
+    Parameters
+    ----------
+    param1 : int
+        First parameter
+    *args : tuple
+        Arguments
+    **kwargs : dict
+        Keyword arguments
+    """, config=cfg)
+    line_by_line_compare(str(doc), r"""
+:Parameters:
+
+    **param1** : int
+        First parameter
+
+    **\*args** : tuple
+        Arguments
+
+    **\*\*kwargs** : dict
+        Keyword arguments
+    """)
+
+def test_autoclass():
+    cfg=dict(show_class_members=True,
+             show_inherited_class_members=True)
+    doc = SphinxClassDoc(str, '''
+A top section before
+
+.. autoclass:: str
+    ''', config=cfg)
+    line_by_line_compare(str(doc), r'''
+A top section before
+
+.. autoclass:: str
+
+.. rubric:: Methods
+
+
+    ''', 5)
+
+
+xref_doc_txt = """
+Test xref in Parameters, Other Parameters and Returns
+
+Parameters
+----------
+p1 : int
+    Integer value
+
+p2 : float, optional
+    Integer value
+
+Other Parameters
+----------------
+p3 : list[int]
+    List of integers
+p4 : :class:`pandas.DataFrame`
+    A dataframe
+p5 : sequence of `int`
+    A sequence
+
+Returns
+-------
+out : array
+    Numerical return value
+"""
+
+
+xref_doc_txt_expected = r"""
+Test xref in Parameters, Other Parameters and Returns
+
+
+:Parameters:
+
+    **p1** : :class:`python:int`
+        Integer value
+
+    **p2** : :class:`python:float`, optional
+        Integer value
+
+:Returns:
+
+    **out** : :obj:`array <numpy.ndarray>`
+        Numerical return value
+
+
+:Other Parameters:
+
+    **p3** : :class:`python:list`\[:class:`python:int`]
+        List of integers
+
+    **p4** : :class:`pandas.DataFrame`
+        A dataframe
+
+    **p5** : :obj:`python:sequence` of `int`
+        A sequence
+"""
+
+
+def test_xref():
+    xref_aliases = {
+        'sequence': ':obj:`python:sequence`',
+    }
+
+    class Config():
+        def __init__(self, a, b):
+            self.numpydoc_xref_aliases = a
+            self.numpydoc_xref_aliases_complete = b
+            # numpydoc.update_config fails if this config option not present
+            self.numpydoc_validation_checks = set()
+            self.numpydoc_validation_exclude = set()
+
+    xref_aliases_complete = deepcopy(DEFAULT_LINKS)
+    for key in xref_aliases:
+        xref_aliases_complete[key] = xref_aliases[key]
+    config = Config(xref_aliases, xref_aliases_complete)
+    app = namedtuple('config', 'config')(config)
+    update_config(app)
+
+    xref_ignore = {'of', 'default', 'optional'}
+
+    doc = SphinxDocString(
+        xref_doc_txt,
+        config=dict(
+            xref_param_type=True,
+            xref_aliases=xref_aliases_complete,
+            xref_ignore=xref_ignore
+        )
+    )
+
+    line_by_line_compare(str(doc), xref_doc_txt_expected)
 
 
 if __name__ == "__main__":
-    import nose
-    nose.run()
+    import pytest
+    pytest.main()
